'use strict'
var __importDefault =
  (this && this.__importDefault) ||
  function (mod) {
    return mod && mod.__esModule ? mod : { default: mod }
  }
Object.defineProperty(exports, '__esModule', { value: true })
const path_1 = __importDefault(require('path'))
const log = require('debug')('bp:worker')
const webpack_1 = __importDefault(require('webpack'))
const memory_fs_1 = __importDefault(require('memory-fs'))
const is_valid_npm_name_1 = __importDefault(require('is-valid-npm-name'))
const zlib_1 = require('zlib')
const fs_1 = __importDefault(require('fs'))
const getDependencySizeTree_1 = __importDefault(
  require('../getDependencySizeTree')
)
const getParseTime_1 = __importDefault(require('../getParseTime'))
const makeWebpackConfig_1 = __importDefault(
  require('../config/makeWebpackConfig')
)
const perf_hooks_1 = require('perf_hooks')
const CustomError_1 = require('../errors/CustomError')
const telemetry_utils_1 = __importDefault(require('./telemetry.utils'))
const BuildUtils = {
  createEntryPoint(packageName, installPath, options) {
    const entryPath = path_1.default.join(
      installPath,
      options.entryFilename || 'index.js'
    )
    let importStatement
    if (options.esm) {
      if (options.customImports) {
        importStatement = `
          import { ${options.customImports.join(', ')} } from '${packageName}'; 
          console.log(${options.customImports.join(', ')})
     `
      } else {
        importStatement = `import p from '${packageName}'; console.log(p)`
      }
    } else {
      if (options.customImports) {
        importStatement = `
        const { ${options.customImports.join(
          ', '
        )} } = require('${packageName}'); 
        console.log(${options.customImports.join(', ')})
        `
      } else {
        importStatement = `const p = require('${packageName}'); console.log(p)`
      }
    }
    try {
      fs_1.default.writeFileSync(entryPath, importStatement, 'utf-8')
      return entryPath
    } catch (err) {
      throw new CustomError_1.EntryPointError(err)
    }
  },
  compilePackage({ name, entry, externals, debug, minifier }) {
    const startTime = perf_hooks_1.performance.now()
    const compiler = webpack_1.default(
      makeWebpackConfig_1.default({
        packageName: name,
        entry,
        externals,
        debug,
        minifier,
      })
    )
    const memoryFileSystem = new memory_fs_1.default()
    compiler.outputFileSystem = memoryFileSystem
    return new Promise(resolve => {
      compiler.run((err, stats) => {
        const error = err // Webpack types incorrect
        // stats object can be empty if there are build errors
        resolve({ stats, error, memoryFileSystem })
        if (error) {
          console.error(error)
          telemetry_utils_1.default.compilePackage(
            name,
            false,
            startTime,
            { minifier },
            error
          )
        } else {
          telemetry_utils_1.default.compilePackage(name, true, startTime, {
            minifier,
          })
        }
      })
    })
  },
  _parseMissingModules(errors) {
    const missingModuleErrors = errors.filter(
      error => error.name === 'ModuleNotFoundError'
    )
    if (!missingModuleErrors.length) {
      return []
    }
    // There's a better way to get the missing module's name, maybe ?
    const missingModuleRegex = /Can't resolve '(.+)' in/
    const missingModules = missingModuleErrors.map(err => {
      const matches = err.error.toString().match(missingModuleRegex)
      if (!matches) {
        throw new CustomError_1.UnexpectedBuildError(
          'Expected to find a file path in the module not found error, but found none. Regex for this might be out of date.'
        )
      }
      const missingFilePath = matches[1]
      let packageNameMatch
      if (missingFilePath.startsWith('@')) {
        packageNameMatch = missingFilePath.match(/@[^\/]+\/[^\/]+/) // @babel/runtime/object/create -> @babel/runtime
      } else {
        packageNameMatch = missingFilePath.match(/[^\/]+/) // babel-runtime/object/create -> babel-runtime
      }
      if (!packageNameMatch) {
        throw new CustomError_1.UnexpectedBuildError(
          'Failed to resolve the missing package name. Regex for this might be out of date.'
        )
      }
      return packageNameMatch[0]
    })
    let uniqueMissingModules = Array.from(new Set(missingModules))
    uniqueMissingModules = uniqueMissingModules.filter(
      mod => !mod.startsWith(`${uniqueMissingModules[0]}/`)
    )
    return uniqueMissingModules
  },
  async buildPackage({ name, installPath, externals, options }) {
    var _a
    let entry = {}
    if (options.splitCustomImports) {
      if (!options.customImports || !options.customImports.length) {
        return { assets: [] }
      }
      options.customImports.forEach(importt => {
        entry[importt] = BuildUtils.createEntryPoint(name, installPath, {
          customImports: [importt],
          entryFilename: importt,
          esm: true,
        })
      })
    } else {
      entry['main'] = BuildUtils.createEntryPoint(name, installPath, {
        esm: false,
        customImports: options.customImports,
      })
    }
    log('build start %s', name)
    const { stats, error, memoryFileSystem } = await BuildUtils.compilePackage({
      name,
      entry,
      externals,
      debug: options.debug,
      minifier: options.minifier,
    })
    log('build end %s', name)
    const jsonStatsStartTime = perf_hooks_1.performance.now()
    let jsonStats = stats.toJson({
      assets: true,
      children: false,
      chunks: false,
      chunkGroups: false,
      chunkModules: false,
      chunkOrigins: false,
      modules: true,
      errorDetails: false,
      entrypoints: false,
      reasons: false,
      maxModules: 500,
      performance: false,
      source: true,
      depth: true,
      providedExports: true,
      warnings: false,
      modulesSort: 'depth',
    })
    if (!jsonStats) {
      telemetry_utils_1.default.parseWebpackStats(
        name,
        false,
        jsonStatsStartTime
      )
      throw new CustomError_1.UnexpectedBuildError(
        'Expected webpack json stats to be non-null, but was null'
      )
    } else {
      telemetry_utils_1.default.parseWebpackStats(
        name,
        true,
        jsonStatsStartTime
      )
    }
    if (error && !stats) {
      throw new CustomError_1.BuildError(error)
    } else if (stats.compilation.errors && stats.compilation.errors.length) {
      const missingModules = BuildUtils._parseMissingModules(
        stats.compilation.errors
      )
      if (missingModules.length) {
        if (missingModules.length === 1 && missingModules[0] === name) {
          throw new CustomError_1.EntryPointError(
            stats.compilation.errors.map(err => err.toString())
          )
        } else {
          throw new CustomError_1.MissingDependencyError(
            stats.compilation.errors.map(err => err.toString()),
            { missingModules }
          )
        }
      } else if (jsonStats.errors && jsonStats.errors.length > 0) {
        if (
          jsonStats.errors.some(error =>
            error.includes("Unexpected character '#'")
          )
        ) {
          throw new CustomError_1.CLIBuildError(jsonStats.errors)
        } else {
          throw new CustomError_1.BuildError(jsonStats.errors)
        }
      } else {
        throw new CustomError_1.UnexpectedBuildError(
          'The webpack stats object was unexpectedly empty'
        )
      }
    } else {
      const getAssetStats = asset => {
        const bundle = path_1.default.join(process.cwd(), 'dist', asset.name)
        const bundleContents = memoryFileSystem.readFileSync(bundle)
        let parseTimes = null
        if (options.calcParse) {
          parseTimes = getParseTime_1.default(bundleContents)
        }
        const gzip = zlib_1.gzipSync(bundleContents, {}).length
        const matches = asset.name.match(/(.+?)\.bundle\.(.+)$/)
        if (!matches) {
          throw new CustomError_1.UnexpectedBuildError(
            'Found an asset without the `.bundle` suffix. ' +
              'A loader customization might be needed to recognize this asset type' +
              asset.name
          )
        }
        const [, entryName, extension] = matches
        return {
          name: entryName,
          type: extension,
          size: asset.size,
          gzip,
          parse: parseTimes,
        }
      }
      const assetsGzipStartTime = perf_hooks_1.performance.now()
      const assetStats =
        (_a =
          jsonStats === null || jsonStats === void 0
            ? void 0
            : jsonStats.assets) === null || _a === void 0
          ? void 0
          : _a
              .filter(asset => !asset.chunkNames.includes('runtime'))
              .filter(asset => !asset.name.endsWith('LICENSE.txt'))
              .map(getAssetStats)
      telemetry_utils_1.default.assetsGZIPParseTime(name, assetsGzipStartTime)
      log('build result %O', assetStats)
      return Object.assign(
        { assets: assetStats || [] },
        options.includeDependencySizes && {
          dependencySizes: await getDependencySizeTree_1.default(
            name,
            jsonStats,
            options.minifier
          ),
        }
      )
    }
  },
  async buildPackageIgnoringMissingDeps({
    name,
    externals,
    installPath,
    options,
  }) {
    const buildStartTime = perf_hooks_1.performance.now()
    let buildIteration = 1
    try {
      const buildResult = await BuildUtils.buildPackage({
        name,
        externals,
        installPath,
        options,
      })
      telemetry_utils_1.default.buildPackage(
        name,
        true,
        buildStartTime,
        Object.assign(Object.assign({}, options), { buildIteration })
      )
      return buildResult
    } catch (e) {
      buildIteration++
      if (
        e instanceof CustomError_1.MissingDependencyError &&
        e.missingModules.length <= 6 &&
        e.missingModules.every(mod => is_valid_npm_name_1.default(mod))
      ) {
        const { missingModules } = e.extra
        const newExternals = Object.assign(Object.assign({}, externals), {
          externalPackages: externals.externalPackages.concat(missingModules),
        })
        log(
          '%s has missing dependencies, rebuilding without %o',
          name,
          missingModules
        )
        const rebuiltResult = await BuildUtils.buildPackage({
          name,
          externals: newExternals,
          installPath,
          options,
        })
        telemetry_utils_1.default.buildPackage(
          name,
          true,
          buildStartTime,
          Object.assign(Object.assign({}, options), {
            buildIteration,
            missingModules,
          })
        )
        return Object.assign(
          { ignoredMissingDependencies: missingModules },
          rebuiltResult
        )
      } else {
        telemetry_utils_1.default.buildPackage(
          name,
          false,
          buildStartTime,
          Object.assign(Object.assign({}, options), { buildIteration }),
          e
        )
        throw e
      }
    }
  },
}
exports.default = BuildUtils
